#!/usr/bin/env python

"""
Simplified program data.

Copyright (C) 2006, 2007 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from simplify.simplified.utils import Structure, WithName, name
from simplify.simplified.program import Subprogram

# Special non-program nodes.

class _Class(Structure, WithName):

    """
    The basis of a Python class. Classes with specific instantiation behaviour
    should inherit from this class.
    """

    def __init__(self, *args, **kw):
        Structure.__init__(self, *args, **kw)
        WithName.__init__(self)

    def full_name(self):
        return "class %s" % self._full_name

    # Utility methods.

    def get_instance_attribute_names(self):

        "Return all attribute names used by the instances of this class."

        names = set()
        for instance in self.instances.values():
            for name in instance.namespace.keys():
                names.add(name)
        return names

    def get_names_to_instances(self):

        """
        Return a tuple containing a mapping from names to instances, and a list
        of sorted instance names.
        """

        d = {}
        names = []
        for instance in self.instances.values():
            name = instance.full_name()
            names.append(name)
            d[name] = instance
        names.sort()
        return d, names

    def get_distinct_instances(self):

        """
        Return a list of instances whose instance attribute types are distinct.
        """

        instances = []
        names_found = []

        # Rather than use the instances directly, get them in name order in
        # order to favour those earlier according to the sorting.

        names_to_instances, instance_names = self.get_names_to_instances()

        for instance_name in instance_names:
            instance = names_to_instances[instance_name]
            names = instance.namespace.names
            if not names in names_found:
                instances.append(instance)
                names_found.append(names)
                
        return instances

class SingleInstanceClass(_Class):

    "A Python class producing only one instance."

    def __init__(self, *args, **kw):
        _Class.__init__(self, *args, **kw)
        self.instance = None

    def has_instance(self, node):
        return self.instance is not None

    def add_instance(self, node, instance):
        self.instance = instance

    def get_instance(self, node):
        return self.instance

    def get_instance_name(self, instance):
        return self._full_name

    # Attribute propagation.

    def get_attribute_for_instance(self, attribute, instance):
        return attribute

class MultipleInstanceClass(_Class):

    "A Python class producing many instances."

    def __init__(self, *args, **kw):
        _Class.__init__(self, *args, **kw)
        self.instances = {}
        self.attributes_for_instances = {}

    def _get_key(self, node):
        return getattr(node, "original", None) # self.module.original

    def has_instance(self, node):
        return self.instances.has_key(self._get_key(node))

    def add_instance(self, node, instance):
        self.instances[self._get_key(node)] = instance

    def get_instance(self, node):
        return self.instances[self._get_key(node)]

    def get_instance_name(self, instance):
        return name(instance, self._full_name)

    # Attribute propagation.

    def get_attribute_for_instance(self, attribute, instance):

        # Create specialised methods.

        if isinstance(attribute.type, Subprogram):
            subprogram = attribute.type

            # Each instance may have its own version of the subprogram.

            key = (subprogram, instance)
            if not self.attributes_for_instances.has_key(key):
                new_subprogram = subprogram.copy(instance, subprogram.full_name())
                subprogram.module.simplifier.subnames[new_subprogram.full_name()] = new_subprogram
                self.attributes_for_instances[key] = Attribute(attribute.context, new_subprogram)
                print "New subprogram", new_subprogram, "for", key

            return self.attributes_for_instances[key]

        # The original nodes are returned for other attributes.

        else:
            return attribute

class SelectiveMultipleInstanceClass(MultipleInstanceClass):

    "A Python class which provides multiple instances depending on the class."

    def _get_key(self, node):
        if self.namespace.has_key("__atomic__"):
            return self
        else:
            return MultipleInstanceClass._get_key(self, node)

class ProlificMultipleInstanceClass(MultipleInstanceClass):

    """
    A Python class which provides multiple instances for different versions of
    methods. In order to avoid unbounded instance production (since new
    instances cause new copies of methods which in turn would cause new
    instances), a relations dictionary is maintained which attempts to map
    "requesting instances" to existing instances, suggesting such instances in
    preference to new ones.
    """

    def __init__(self, *args, **kw):
        MultipleInstanceClass.__init__(self, *args, **kw)
        self.instance_relations = {}

    def _get_key(self, node):
        if self.namespace.has_key("__atomic__"):
            return self
        else:
            return node

    def has_instance(self, node):
        requesting_instance = getattr(node, "instance", None)
        #return requesting_instance is not None and requesting_instance.get_class() is self or \
        return self.instance_relations.has_key(requesting_instance) or self.instances.has_key(self._get_key(node))

    def add_instance(self, node, instance):
        requesting_instance = getattr(node, "instance", None)
        print "New instance", instance, "for", id(node), requesting_instance
        self.instances[self._get_key(node)] = instance
        if requesting_instance is not None:
            self.instance_relations[requesting_instance] = instance
            requesting_instance.get_class().instance_relations[instance] = requesting_instance

    def get_instance(self, node):
        requesting_instance = getattr(node, "instance", None)
        #if requesting_instance is not None and requesting_instance.get_class() is self:
        #    return requesting_instance
        return self.instance_relations.get(requesting_instance) or self.instances[self._get_key(node)]

class Instance(Structure):

    "An instance."

    def full_name(self):
        return self.get_class().get_instance_name(self)

    def get_class(self):
        for n in self.namespace.load("__class__"):
            return n.type
        else:
            raise ValueError, "__class__"

class Constant:

    "A constant initialised with a type name for future processing."

    def __init__(self, name, value):
        self.name = name
        self.value = value
        self.typename = self.value.__class__.__name__

class Attribute:

    """
    An attribute abstraction, indicating the type of the attribute along with
    its context or origin.
    """

    def __init__(self, context, type):
        self.context = context
        self.type = type

    def __repr__(self):
        return "Attribute(%s, %s)" % (repr(self.context), repr(self.type))

    def __eq__(self, other):
        return hasattr(other, "type") and other.type == self.type or other == self.type

    def __hash__(self):
        return id(self.type)

# vim: tabstop=4 expandtab shiftwidth=4
